unit ROOTS; {Unit 6 of BLIMDOS.PAS}

interface

  uses
    GLOBALS, MATHS, UTILS;

  procedure QuadraticDeflation (a: vector;
                  delta0, delta1: real;
                  p: integer;
                  var b: vector;
                  var c0, c1, c2: real);

  function Convergence (b: vector): boolean;

  procedure Bairstow (a: vector;
                  p: integer;
                  var delta0, delta1: real;
                  var b: vector);

  procedure RootsOfFactor (i: integer;
                  delta0, delta1: real;
                  var lambda: complexVector);

  procedure MultiBairstow (p: integer;
                  var alpha: vector;
                  var lambda: complexVector);

  procedure JuryCriterion (alpha: vector;
                  p: integer;
                  var stable: boolean);

  procedure CholeskyCriterion (alpha: vector;
                  p: integer;
                  var stable: boolean);

  procedure ComplexPoly (a: complexVector;
                  p: integer;
                  z: complex;
                  var g0: complex;
                  var b: complexVector);

  procedure CpolyValue (a: complexVector;
                  p: integer;
                  z: complex;
                  var f: complex;
                  var b: complexVector);

  procedure Mueller (p: integer;
                  poly: complexvector;
                  var root: complex;
                  var quotient: complexVector);

  procedure ComplexRoots (p: integer;
                  var a, b, lambda: complexVector);

  procedure FindRoots (var p: integer;
                  var complexAlpha, cartRoot: complexVector;
                  var polarRoot: polarVector);

  procedure FindPolesAndZeros (var transferFunct: ARMARecord);

implementation

{_________________________________________________}

  procedure QuadraticDeflation (a: vector;
                  delta0, delta1: real;
                  p: integer;
                  var b: vector;
                  var c0, c1, c2: real);

    var
      i: integer;

  begin {QuadraticDeflation}

    b[p - 2] := a[p];
    b[p - 3] := a[p - 1] - delta1 * b[p - 2];
    c1 := 0;
    c2 := -b[p - 2];

    for i := 4 to p + 2 do
      begin
        b[p - i] := a[p - i + 2] - delta1 * b[p - i + 1] - delta0 * b[p - i + 2];
        c0 := c1;
        c1 := c2;
        c2 := -b[p - i + 1] - delta1 * c1 - delta0 * c0;
      end;

  end; {QuadraticDeflation}

   {_________________________________________________}

  function Convergence (b: vector): boolean;

  begin
    if (Abs(b[-1]) < 1E-6) and (Abs(b[-2]) < 1E-6) then
      convergence := true
    else
      convergence := false;
  end; {Convergence}

   {_________________________________________________}

  procedure Bairstow (a: vector;
                  p: integer;
                  var delta0, delta1: real;
                  var b: vector);

    var
      iterations: integer;
      c0, c1, c2, det: real;
      convergence: boolean;

  begin {Bairstow}
    iterations := 0;
    convergence := false;

    repeat
      QuadraticDeflation(a, delta0, delta1, p, b, c0, c1, c2);
      det := Sqr(c1) - c0 * c2;
      delta1 := delta1 - (c1 * b[-1] - c0 * b[-2]) / det;
      delta0 := delta0 - (c1 * b[-2] - c2 * b[-1]) / det;
      iterations := iterations + 1;
      if (Abs(b[-1]) < 1E-5) and (Abs(b[-2]) < 1E-5) then
        convergence := true;
    until (convergence) or (iterations > 30);

    if not convergence then
      Writeln('the program failed to converge');

  end;{Bairstow}

   {_________________________________________________}

  procedure RootsOfFactor (i: integer;
                  delta0, delta1: real;
                  var lambda: complexVector);

  begin

    if Sqr(delta1) <= 4 * delta0 then
      begin {complex roots}
        lambda[i].re := -delta1 / 2;
        lambda[i - 1].re := -delta1 / 2;
        lambda[i].im := Sqrt(4 * delta0 - Sqr(delta1)) / 2;
        lambda[i - 1].im := -Sqrt(4 * delta0 - Sqr(delta1)) / 2;
        Writeln(' complex conjugate roots');
        Writeln(' real part =', lambda[i].re : 5 : 5, '   imaginary part = +/-', lambda[i].im : 5 : 5);
        Writeln;
      end
    else if sqr(delta1) > 4 * delta0 then
      begin {real roots}
        lambda[i].re := (-delta1 + Sqrt(sqr(delta1) - 4 * delta0)) / 2;
        lambda[i - 1].re := (-delta1 - Sqrt(sqr(delta1) - 4 * delta0)) / 2;
        lambda[i].im := 0.0;
        lambda[i - 1].im := 0.0;
        Writeln('first real root = ', lambda[i].re : 5 : 5);
        Writeln('second real = ', lambda[i - 1].re : 5 : 5);
        Writeln;
      end;

  end; {RootsOfFactor}

   {_________________________________________________}

  procedure MultiBairstow (p: integer;
                  var alpha: vector;
                  var lambda: complexVector);

    var
      i, j, r: integer;
      c0, c1, c2, delta0, delta1: real;
      beta, quotient: vector;

  begin {MultiBairstow}
    if odd(p) then
      r := (p - 1) div 2
    else
      r := p div 2;

    beta := alpha;

    for i := 0 to r - 1 do
      begin
        j := p - 2 * i;
        delta1 := 0;
        delta0 := 0;
        Bairstow(beta, j, delta0, delta1, quotient); {Initial value of factor}
        Bairstow(alpha, p, delta0, delta1, quotient); {refined value}
        RootsOfFactor(j, delta0, delta1, lambda);
        QuadraticDeflation(beta, delta0, delta1, p, quotient, c0, c1, c2);
        beta := quotient;
      end;
    if odd(p) then
      begin
        lambda[1].re := -quotient[0];
        lambda[1].im := 0.0
      end;

  end; {MultiBairstow}

   {_________________________________________________}

  procedure JuryCriterion (alpha: vector;
                  p: integer;
                  var stable: boolean);
    var
      i, n, fact: integer;
      Q, a0, an, ai, anmi: real;

  begin {Jury Criterion }
    stable := true;

    {Preliminary Testing}
    if Abs(alpha[0]) <= Abs(alpha[p]) then
      begin
        stable := false;
        Writeln('Abs(alpha[0]) <= Abs(alpha[p]): stable is false')
      end;

    q := alpha[0];
    for i := 1 to p do
      q := q + alpha[i];
    if q <= 0 then
      begin
        stable := false;
        Writeln('q(1) <= 0: stable is false')
      end;

    q := alpha[0];
    fact := 1;
    for i := 1 to p do
      begin
        fact := -1 * fact;
        q := q + fact * alpha[i]
      end;

    if q <= 0 then
      begin
        stable := false;
        Writeln(' q(-1) <= 0: stable is false')
      end;

    if (stable = true) and (p > 2) then
      begin {Further Testing}

        Writeln('    The polynomial has passed the preliminary tests');
        Writeln;

        n := p;
        repeat
          begin {repeat}
            a0 := alpha[0];
            an := alpha[n];
            alpha[0] := a0 * a0 - an * an;
            for i := 1 to n div 2 do
              begin {i}
                anmi := alpha[n - i];
                ai := alpha[i];
                alpha[i] := a0 * ai - an * anmi;
                alpha[n - i] := a0 * anmi - an * ai;
              end; {i}

            if Abs(alpha[0]) <= Abs(alpha[n - 1]) then
              begin
                stable := false;
                Writeln('    THE POLYNOMIAL IS UNSTABLE');
                Writeln;
              end;
            n := n - 1;
          end; {repeat}
        until (n = 2);

      end;{further testing}

  end; {Jury Criterion}

    {_________________________________________________}

  procedure CholeskyCriterion (alpha: vector;
                  p: integer;
                  var stable: boolean);

    var
      i, j, k: integer;
      det: real;
      q: matrix;

  begin {Cholesky Criterion}
    stable := true;

    {Find the Inverse Disperion matrix}
    for i := 0 to p - 1 do
      for j := 0 to i do
        begin {i,j}
          q[i, j] := 0.0;
          for k := 0 to p - 1 - i do
            begin {k}
              q[i, j] := q[i, j] + alpha[k] * alpha[k + i - j];
              q[i, j] := q[i, j] - alpha[p - k] * alpha[p - k - i + j];
            end; {k}
        end;{i,j}

    {Factorise the matrix}
    det := 1;
    for i := 0 to p - 1 do
      begin {i}
        for j := 0 to i do
          begin {j}
            for k := 0 to j - 1 do
              q[i, j] := q[i, j] - q[k, k] * q[i, k] * q[j, k];
            if i > j then
              q[i, j] := q[i, j] / q[j, j];
          end; {j}
        det := det * q[i, i];
        if det <= 0 then
          stable := false;
      end; {i}

  end; {Cholesky Criterion}

 {_________________________________________________}

  procedure ComplexPoly (a: complexVector;
                  p: integer;
                  z: complex;
                  var g0: complex;
                  var b: complexvector);

    var
      i: integer;

  begin
    b[p - 1] := a[p];
    for i := 1 to p - 1 do
      begin
        ComplexMult(b[p - i], z, b[p - i - 1]);
        ComplexAdd(b[p - i - 1], a[p - i], b[p - i - 1]);
      end;
    ComplexMult(b[0], z, g0);
    ComplexAdd(g0, a[0], g0);
  end; {ComplexPoly}

   {_________________________________________________}

  procedure CpolyValue (a: complexVector;
                  p: integer;
                  z: complex;
                  var f: complex;
                  var b: complexVector);

    var
      i: integer;

  begin {CpolyValue}
    b[p - 1].re := a[p].re;
    b[p - 1].im := a[p].im;
    for i := 1 to p - 1 do
      begin
        b[p - i - 1].re := b[p - i].re * z.re - b[p - i].im * z.im + a[p - i].re;
        b[p - i - 1].im := b[p - i].im * z.re + b[p - i].re * z.im + a[p - i].im;
      end;
    f.re := b[0].re * z.re - b[0].im * z.im + a[0].re;
    f.im := b[0].im * z.re + b[0].re * z.im + a[0].im;
  end; {CpolyValue}

{_________________________________________________}

  procedure Mueller (p: integer;
                  poly: complexVector;
                  var root: complex;
                  var quotient: complexVector);

    const
      almostZero = 1E-15;

    var
      iterations, exeat: integer;
      convergence: boolean;
      a, b, c, z0, z1, z2, z3, h, h1, h2, h3, f0, f1, f2, f3: complex;
      delta1, delta2, discrim, denom, store1, store2: complex;

  begin{Mueller}

       {set the intial values}
    z2.re := -0.75 * root.re - 1;
    z2.im := root.im;
    z1.re := 0.75 * root.re;
    z1.im := 1.2 * root.im + 1;
    z0 := root;
    ComplexPoly(poly, p, z0, f0, quotient);
    ComplexPoly(poly, p, z1, f1, quotient);
    ComplexPoly(poly, p, z2, f2, quotient);

    iterations := 0;
    exeat := 0;
    convergence := false;

    while (not convergence) and (exeat = 0) and (iterations < 60) do
      begin
        ComplexSub(f1, f0, delta1);
        ComplexSub(f2, f0, delta2);
        ComplexSub(z1, z0, h1);
        ComplexSub(z2, z0, h2);
        ComplexSub(z1, z2, h3);
        ComplexMult(h1, h2, h);
        ComplexMult(h, h3, h);
        if Cmod(h) < almostZero then
          exeat := 1;      {Can't fit a quadratic to these points}

        if exeat = 0 then
          begin  {calculate coefficients of the quadratic}
            ComplexMult(h2, delta1, store1);
            ComplexMult(h1, delta2, store2);
            ComplexSub(store1, store2, a);
            ComplexDiv(a, h, a); { a = (h_2\delta_1 - h_1\delta_2)/h }
            ComplexMult(store1, h2, store1);
            ComplexMult(store2, h1, store2);
            ComplexSub(store2, store1, b);
            ComplexDiv(b, h, b); { b = (h_1^2\delta_2 - h_2^2\delta_1)/h }
            c := f0;
          end;

        if (Cmod(a) <= almostZero) and (Cmod(b) <= almostZero) then
          exeat := 2;     {Test if parabola is really a constant }

        ComplexMult(b, b, discrim);  { b^2 }
        ComplexMult(a, c, h);
        discrim.re := discrim.re - 4 * h.re;  { b^2 - 4ac }
        discrim.im := discrim.im - 4 * h.im;
        ComplexSqrt(discrim, discrim); { Sqrt( b^2 - 4ac) }

        if (discrim.re * b.re + discrim.im * b.im > 0) then
          ComplexAdd(b, discrim, denom)
        else
          ComplexSub(b, discrim, denom);

        if Cmod(denom) < almostZero then
          begin  { if b \pm \Sqrt(b^2 - 4ac) \simeq 0 }
            z3.re := 0;
            z3.im := 0;
          end
        else
          begin
            ComplexAdd(c, c, h);
            ComplexDiv(h, denom, h);
            ComplexSub(z0, h, z3);
          end;

        ComplexPoly(poly, p, z3, f3, quotient);
        if (Cmod(h) < 1E-10) or (Cmod(f3) < 1E-10) then
          convergence := true;

        z2 := z1;
        f2 := f1;
        z1 := z0;
        f1 := f0;
        z0 := z3;
        f0 := f3;

        iterations := iterations + 1;
      end; {while}
    root := z3;

{if not convergence then}
{Writeln('    The program failed to converge');}

{Writeln('    root.re  = ', root.re : 8 : 4);}
{Writeln('    root.im  = ', root.im : 8 : 4);}
{Writeln('    Exit = ', exeat : 1);}
{Writeln('    iterations = ', iterations : 5);}
{Writeln}

  end; {Mueller}

   {_________________________________________________}

  procedure ComplexRoots (p: integer;
                  var a, b, lambda: complexVector);

    var
      i: integer;
      root: complex;
      q: complexVector;

  begin {ComplexRoots}

    b := a;
    for i := 0 to p - 2 do
      begin
        root.re := 0;
        root.im := 0;
        Mueller(p - i, b, root, q);
        lambda[i + 1] := root;
        b := q;
      end;
    lambda[p].re := -b[0].re / a[p].re;
    lambda[p].im := -b[0].im / a[p].re;

  end; {ComplexRoots}

{________________________________________________}

  procedure FindRoots (var p: integer;
                  var complexAlpha, cartRoot: complexVector;
                  var polarRoot: polarVector);

{This procedure reverses the order of the coefficients in order to place the stable roots}
{inside the unit circle. It finds both the Cartesian form and the polar formof the roots.}

    const
      virtualZero = 1E-10;

    var
      i: integer;
      pi, radToDeg: real;
      Calpha, Cbeta: complexVector;

  begin
    pi := 4 * ArcTan(1.0);
    radToDeg := 180.0 / pi;

    for i := 0 to p do
      begin
        Calpha[i].re := complexAlpha[p - i].re;
        Calpha[i].im := 0.0;
      end;
    ComplexRoots(p, Calpha, Cbeta, cartRoot);

    for i := 1 to p do
      begin
        polarRoot[i].modu := Sqrt(Sqr(cartRoot[i].re) + Sqr(cartRoot[i].im));
        if Abs(cartRoot[i].re) < virtualZero then
          begin {if}
            if cartRoot[i].im < 0 then
              polarRoot[i].argu := pi / 2
            else
              polarRoot[i].argu := -pi / 2
          end {if}
        else if cartRoot[i].re < 0 then
          polarRoot[i].argu := ArcTan(cartRoot[i].im / cartRoot[i].re) + pi
        else
          polarRoot[i].argu := ArcTan(cartRoot[i].im / cartRoot[i].re);
        polarRoot[i].argu := polarRoot[i].argu * radToDeg;
      end; {i}

  end; {FindRoots}

{_________________________________________________}

  procedure FindPolesAndZeros (var transferFunct: ARMARecord);

    var
      i, p, q: integer;
      Cgamma, Cbeta, poles, zeros: complexVector;
 {Why not surround this with with transferFunct statement}

  begin
    p := transferFunct.p;
    q := transferFunct.q;

       {Writeln(' THE POLES OF THE TRANSFER FUNCTION'); }
       {READLN;}

    for i := 0 to p do
      begin
        Cgamma[i].re := transferFunct.alpha[p - i];
        Cgamma[i].im := 0.0;
      end;
    ComplexRoots(p, Cgamma, Cbeta, poles); {Poles}
    for i := p downto 1 do
      begin
        transferFunct.poles[i].re := poles[i].re;
        transferFunct.poles[i].im := poles[i].im;
{modulus := Sqrt(poles[i].re * poles[i].re + poles[i].im * poles[i].im);}
{Write('    poles[', i : 1, '].re = ', poles[i].re : 5 : 5, '   poles[', i : 1, '].im = ', poles[i].im : 5 : 5);}
{Writeln('    modulus = ', modulus : 5 : 5);}
      end;

{Writeln;}
{Writeln(' ' : 60, '<RETURN>');}
{Readln;}

    if q > 1 then
      begin
        for i := 0 to q do
          begin
            Cgamma[i].re := transferFunct.mu[q - i];
            Cgamma[i].im := 0.0;
          end;
        ComplexRoots(q, Cgamma, Cbeta, zeros); {Zeros}
        for i := q downto 1 do
          begin {i}
            transferFunct.zeros[i].re := zeros[i].re;
            transferFunct.zeros[i].im := zeros[i].im;
          end; {i}
      end; {if}

    transferFunct.rootsFound := true;

  end; {FindPolesAndZeros}

{________________________________________________}

end. {ROOTS: Unit 6 of BLIMDOS.PAS}

